#include "BoundingBox.h"

#include "../Triangle/Triangle.h"
#include "../Vertex/Vertex.h"
#include <GridGenerator/utilities/math/Math.h>

#include <limits>



 BoundingBox::BoundingBox(real minX, real maxX, real minY, real maxY, real minZ, real maxZ) : minX(minX), maxX(maxX), minY(minY), maxY(maxY), minZ(minZ), maxZ(maxZ) {}

 HOSTDEVICE BoundingBox::BoundingBox() :
     minX(0),
     maxX(0),
     minY(0),
     maxY(0),
     minZ(0),
     maxZ(0) {}

 BoundingBox::BoundingBox(const BoundingBox &t) : minX(t.minX), maxX(t.maxX), minY(t.minY), maxY(t.maxY), minZ(t.minZ), maxZ(t.maxZ) {}


 CUDA_HOST BoundingBox BoundingBox::makeInvalidMinMaxBox()
 {
     BoundingBox box = BoundingBox(std::numeric_limits<real>::max(),
         std::numeric_limits<real>::lowest(),
         std::numeric_limits<real>::max(),
         std::numeric_limits<real>::lowest(),
         std::numeric_limits<real>::max(),
         std::numeric_limits<real>::lowest());
     return box;
 }

 void BoundingBox::setMinMax(const Triangle& t)
 {
     real minX, maxX, minY, maxY, minZ, maxZ;
     t.setMinMax(minX, maxX, minY, maxY, minZ, maxZ);
     if(minX < this->minX)
         this->minX = minX;
     if (minY < this->minY)
         this->minY = minY;
     if (minZ < this->minZ)
         this->minZ = minZ;

     if (maxX > this->maxX)
         this->maxX = maxX;
     if (maxY > this->maxY)
         this->maxY = maxY;
     if (maxZ > this->maxZ)
         this->maxZ = maxZ;
 }


 bool BoundingBox::intersect(const Triangle &t) const
 {
	 if (isInside(t.v1) || isInside(t.v2) || isInside(t.v3))
		 return true;
	 return false;
 }


 bool BoundingBox::isInside(const Triangle &t) const
 {
	 if (isInside(t.v1) && isInside(t.v2) && isInside(t.v3))
		 return true;
	 return false;
 }

 bool BoundingBox::isInside(const real x, const real y, const real z) const
 {
     return this->isInside(Vertex(x,y,z));
 }

 bool BoundingBox::isInside(const Vertex &v) const
 {
     if (v.isXbetween(minX, maxX) && v.isYbetween(minY, maxY) && v.isZbetween(minZ, maxZ))
		 return true;
	 return false;
 }

 std::vector<std::vector<Vertex> > BoundingBox::getIntersectionPoints(const BoundingBox &b) const
 {
	 std::vector<std::vector<Vertex> > intersectionBox;
	 intersectionBox.resize(6);

	 int intersects = 0;
	 if (b.minX < maxX && b.maxX > maxX) { //maxX is intersect
		 intersectionBox[intersects].push_back(Vertex((real)maxX, (real)minY, (real)minZ));
		 intersectionBox[intersects].push_back(Vertex((real)maxX, (real)maxY, (real)minZ));
		 intersectionBox[intersects].push_back(Vertex((real)maxX, (real)minY, (real)maxZ));
		 intersects++;
	 }
	 if (b.minX < minX && b.maxX > minX) { //minX is intersect
		 intersectionBox[intersects].push_back(Vertex((real)minX, (real)minY, (real)minZ));
		 intersectionBox[intersects].push_back(Vertex((real)minX, (real)maxY, (real)minZ));
		 intersectionBox[intersects].push_back(Vertex((real)minX, (real)minY, (real)maxZ));
		 intersects++;
	 }
	 if (b.minY < minY && b.maxY > minY) { //minY is intersect
		 intersectionBox[intersects].push_back(Vertex((real)minX, (real)minY, (real)minZ));
		 intersectionBox[intersects].push_back(Vertex((real)maxX, (real)minY, (real)minZ));
		 intersectionBox[intersects].push_back(Vertex((real)minX, (real)minY, (real)maxZ));
		 intersects++;
	 }
	 if (b.minY < maxY && b.maxY > maxY) { //maxY is intersect
		 intersectionBox[intersects].push_back(Vertex((real)minX, (real)maxY, (real)minZ));
		 intersectionBox[intersects].push_back(Vertex((real)maxX, (real)maxY, (real)minZ));
		 intersectionBox[intersects].push_back(Vertex((real)minX, (real)maxY, (real)maxZ));
		 intersects++;
	 }
	 if (b.minZ < minZ && b.maxZ > minZ) { //minZ is intersect
		 intersectionBox[intersects].push_back(Vertex((real)minX, (real)minY, (real)minZ));
		 intersectionBox[intersects].push_back(Vertex((real)maxX, (real)minY, (real)minZ));
		 intersectionBox[intersects].push_back(Vertex((real)minX, (real)maxY, (real)minZ));
		 intersects++;
	 }
	 if (b.minZ < maxZ && b.maxZ > maxZ) { //maxZ is intersect
		 intersectionBox[intersects].push_back(Vertex((real)minX, (real)minY, (real)maxZ));
		 intersectionBox[intersects].push_back(Vertex((real)maxX, (real)minY, (real)maxZ));
		 intersectionBox[intersects].push_back(Vertex((real)minX, (real)maxY, (real)maxZ));
		 intersects++;
	 }

	 return intersectionBox;
 }

 bool BoundingBox::intersect(const BoundingBox &box) const
 {
	 struct Vertex v[8];
	 box.getPoints(v);

	 for (int i = 0; i < 8; i++) {
		 if (isInside(v[i]))
			 return true;
	 }
	 return false;
 }

 void BoundingBox::getPoints(Vertex v[8]) const
 {
	 v[0] = Vertex(minX, minY, minZ);
	 v[1] = Vertex(maxX, minY, minZ);
	 v[2] = Vertex(minX, maxY, minZ);
	 v[3] = Vertex(maxX, maxY, minZ);

	 v[4] = Vertex(minX, minY, maxZ);
	 v[5] = Vertex(maxX, minY, maxZ);
	 v[6] = Vertex(minX, maxY, maxZ);
	 v[7] = Vertex(maxX, maxY, maxZ);
 }


 void BoundingBox::print() const
 {
	 printf("min/max - x: %2.4f/ %2.4f, y: %2.4f, %2.4f, z: %2.4f, %2.4f \n", minX, maxX, minY, maxY, minZ, maxZ);
 }


 CUDA_HOST bool BoundingBox::operator==(const BoundingBox &box) const
 {
     return vf::Math::equal(minX, box.minX)
         && vf::Math::equal(maxX, box.maxX)
         && vf::Math::equal(minY, box.minY)
         && vf::Math::equal(maxY, box.maxY)
         && vf::Math::equal(minZ, box.minZ)
         && vf::Math::equal(maxZ, box.maxZ);
 }

 void BoundingBox::extend(real delta)
 {
     this->minX -= delta;
     this->minY -= delta;
     this->minZ -= delta;

     this->maxX += delta;
     this->maxY += delta;
     this->maxZ += delta;
 }


